#!/usr/bin/env python

from __future__ import with_statement

from os import system, makedirs
system('find .. -name "*.pyc" | xargs rm -f')

from glob import glob
import sys
for depspath in glob('../deps.d/*'):
    sys.path.insert(0, depspath)
sys.path.insert(0, '..')

from unittest import TestCase, main
from cq2utils import CallTrace, autostart, getRequest, postRequest
from os.path import join, isdir, isfile
from os import remove, rename
from time import sleep, time
from amara.binderytools import bind_file
from merescocore.components.http import ObservableHttpServer
from merescocore.framework import be
from urllib import urlencode
from lxml.etree import tostring

from bzv.server import main as serverMain
from bzv import Inbox
from weightless import Reactor, compose
from socket import socket
from random import randint
from socket import socket
from errno import EAGAIN as TEMPORARILY_UNAVAILABLE

global port
global reactor
port = None
reactor = None
integrationTempdir = '/tmp/bzv/integrationtest'
inboxDirectory = join(integrationTempdir, 'inboxDirectory')
doneDirectory = join(integrationTempdir, 'doneDirectory')

class IntegrationTest(TestCase):

    def testExplain(self):
        header, body = getRequest(reactor, port, '/sru', {}, parse=False)
        self.assertTrue('explainResponse' in body, body)
        self.assertTrue('port>%s</' % port in body, body)

    def testHtml(self):
        header, response = getRequest(reactor, port, '/elsevier', {}, parse=False)
        self.assertTrue('</html>' in response, response)

    def testSruQuery(self):
        message = self._doQuery({'query': 'meta.repository.id exact inboxrepo'})
        dynamiclyInserted = int(str(message.searchRetrieveResponse.numberOfRecords))

        self.assertSruQuery(81 + dynamiclyInserted, 'meresco.exists=true')
        self.assertSruQuery(19, 'meta.repository.id exact neeo')
        self.assertSruQuery(61, 'meta.repository.id exact elsevieroai')
        self.assertSruQuery(1,  'meta.repository.id exact pica')
        self.assertSruQuery(19, 'neeo')

    def testModsRecordSchemaIsDefault(self):
        result = self._doQuery({'query':'meresco.exists=true'})
        self.assertEquals('mods', str(result.searchRetrieveResponse.records.record.recordSchema))
        xmlOutput = result.searchRetrieveResponse.records.record.recordData.xml()
        self.assertTrue('<mods xmlns="http://www.loc.gov/mods/v3"' in xmlOutput, xmlOutput)

    def testDrilldown(self):
        result = self.doDrilldown('meresco.exists=true', 'mods.name.namePart')
        navigator = result.extraResponseData.drilldown.term_drilldown.navigator
        self.assertEquals(1, len(navigator))
        self.assertEquals('mods.name.namePart', str(navigator.name))
        self.assertEquals(10, len(navigator.item))

        result = self.doDrilldown('meresco.exists=true', 'meta.repository.id')
        navigator = result.extraResponseData.drilldown.term_drilldown.navigator
        self.assertEquals(1, len(navigator))
        self.assertEquals('meta.repository.id', str(navigator.name))
        self.assertEquals(3, len(navigator.item))
        self.assertEquals(set(['elsevieroai', 'neeo', 'pica']), set([str(item) for item in navigator.item]))

        result = self.doDrilldown('meresco.exists=true', 'mods.name.type')
        navigator = result.extraResponseData.drilldown.term_drilldown.navigator

        self.assertEquals(5, len(navigator.item))
        self.assertEquals(set(['code', 'corporate', 'family', 'given', 'personal']), set([str(item) for item in navigator.item]))

    def testMultiLevelDrilldownInHtml(self):
        html = self.getPage('/nl/page/search?query=')
        self.assertTrue('&drilldown=' in html, html)
        self.assertFalse('&drilldown=drilldown.norm.date.decennium' in html, html[html.index('&drilldown='):])
        self.assertTrue('&drilldown=drilldown.norm.date.year' in html, html[html.index('&drilldown='):])

    def testNormalizationEmptyDateInXml(self):
        result = self._doQuery({'query': 'meta.upload.id exact "elsevieroai:elsevier:CXT0242A:01988859:00690004:08000608"', 'recordSchema': 'extra'})
        extraXml = result.searchRetrieveResponse.records.record.recordData.extra
        self.assertEquals('', str(extraXml.norm.date))

    def testRss(self):
        headers, body = getRequest(reactor, port, '/rss', {'query': 'meta=neeo'}, parse='lxml')
        items = body.xpath('/rss/channel/item')
        self.assertEquals(15, len(items))
        self.assertEquals(['IntegrationTest BZV'], body.xpath('/rss/channel/title/text()'))

    def testUndatedRecordsGetUnknown(self):
        self.assertSruQuery(6, 'mods.originInfo.dateIssued=unknown')

    def testRateRecord(self):
        ratingFile = join(integrationTempdir, 'storage', 'elsevieroai', 'elsevier:CXT0242A:01988859:00690004:08000608', 'rating')
        isfile(ratingFile) and remove(ratingFile)
        recordId = 'elsevieroai:elsevier:CXT0242A:01988859:00690004:08000608'

        result = self._doQuery({'query': 'meta.upload.id exact "%s"' % recordId, 'recordSchema': 'rating'})
        ratingXml = result.searchRetrieveResponse.records.record.recordData.rating
        self.assertEquals('', ratingXml)

        header,body = postRequest(reactor, port, '/rateRecord', urlencode({'recordId': recordId, 'value': 3}), parse=False)
        self.assertTrue('302 ' in header, header+body)

        result = self._doQuery({'query': 'meta.upload.id exact "%s"' % recordId, 'recordSchema': 'rating'})
        ratingXml = result.searchRetrieveResponse.records.record.recordData.rating
        self.assertEquals('3000', str(ratingXml.averageScore))

    def testDidlIsStored(self):
        result = self.assertSruQuery(19, 'meta=neeo', recordSchema='didl')
        xmlData = result.xml()
        self.assertFalse("'didl')' does not exist." in xmlData, xmlData)
        self.assertEquals('urn:nbn:nl:ui:12-402853', str(result.rootNode.searchRetrieveResponse.records.record.recordData.DIDL.Item.Descriptor.Statement.Identifier))

    def testReindex(self):
        if skipReindexTest:
            print "SKIPPING testReindex"
            return

        header, body = self.doReindex('/reindex')
        self.assertEquals('done', body.split('\n')[-1])

    def testSelectiveReindexing(self):
        if skipReindexTest:
            print "SKIPPING testSelectiveReindexing"
            return

        header, body = self.doReindex('/reindex?partialIdentifier=neeo:oai:wo.uvt.nl')
        identifiers = body.split('\n')[:-1]
        self.assertEquals(19, len(identifiers))

        header, body = self.doReindex('/reindex?partialIdentifier=neeo:oai:wo.uvt.nl:402853')
        identifiers = body.split('\n')[:-1]
        self.assertEquals(1, len(identifiers))
        self.assertEquals(['neeo:oai:wo.uvt.nl:402853'], identifiers)

    def doReindex(self, query):
        sok = socket()
        sok.connect(('localhost', port))
        sok.setblocking(0)
        sok.send('GET %s HTTP/1.0\r\n\r\n' % query)

        response = ''

        stopTheLoop = []
        reactor.addTimer(150, lambda: stopTheLoop.append(True))

        while not stopTheLoop:
            try:
                part = sok.recv(1024)
                if not part:
                    break
                response += part
            except Exception, e:
                if e.args[0] != TEMPORARILY_UNAVAILABLE:
                    raise
            reactor.addTimer(0.01, lambda: None)
            reactor.step()
        return response.split('\r\n\r\n')

    def testInbox(self):
        message = self._doQuery({'query': 'meta.repository.id exact inboxrepo'})
        oldRecordCount = int(str(message.searchRetrieveResponse.numberOfRecords))
        recordIdentifier = 'inboxrepo:identifier:%s' % (oldRecordCount+1)

        inboxDirectory = join(integrationTempdir, 'inboxDirectory')
        open('/tmp/' + recordIdentifier, 'w').write("""<mods xmlns="http://www.loc.gov/mods/v3" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" version="3.3" xsi:schemaLocation="http://www.loc.gov/mods/v3 http://www.loc.gov/standards/mods/v3/mods-3-3.xsd">
        <titleInfo><title>Test record for inbox</title></titleInfo>
        <genre>info:eu-repo/semantics/article</genre>
        <language><languageTerm type="code" authority="iso639-2b">eng</languageTerm></language>
        <originInfo>
            <dateIssued qualifier="inferred">20080525</dateIssued>
        </originInfo>
    </mods>""")

        rename('/tmp/' + recordIdentifier, join(inboxDirectory, recordIdentifier+".record"))
        reactor.step()
        #wheelOfTime(reactor, 2)

        message = self._doQuery({'query': 'meta.repository.id exact inboxrepo'})
        newRecordCount = int(str(message.searchRetrieveResponse.numberOfRecords))
        self.assertEquals(oldRecordCount + 1, newRecordCount)

    def doDrilldown(self, query, drilldownField):
        message = self._doQuery({'query':query, 'x-term-drilldown': drilldownField})
        result = message.searchRetrieveResponse
        return result

    def assertSruQuery(self, numberOfRecords, query, **kwargs):
        queryParams = {'query': query}
        queryParams.update(kwargs)
        message = self._doQuery(queryParams)
        result = message.searchRetrieveResponse
        self.assertEquals(numberOfRecords, int(str(result.numberOfRecords)))
        return result

    def _resultCount(self, query):
        queryParams = {'query': query}
        queryParams.update(kwargs)
        message = self._doQuery(queryParams)
        return int(str(message.searchRetrieveResponse.numberOfRecords))

    def _doQuery(self, arguments):
        queryArguments = {'version': '1.1', 'operation': 'searchRetrieve'}
        queryArguments.update(arguments)
        header, body = getRequest(reactor, port, '/sru', queryArguments)
        return body

    def getPage(self, url, parse=False):
        header, body = getRequest(reactor, port, url, {}, parse=parse)
        return body

def createDatabase():
    allOutput = ""
    recordPacking = 'xml'
    start = time()
    print "Creating database in", integrationTempdir
    sourceFiles = glob('sahara_output/*.updateRequest')
    for updateRequestFile in sorted(sourceFiles):
        print 'Sending:', updateRequestFile
        header, body = postRequest(reactor, port, '/update', open(updateRequestFile).read())
        if '200 Ok' not in header:
            print 'No 200 Ok response, but:'
            print header
            exit(123)
        if "srw:diagnostics" in body.xml():
            print body.xml()
            exit(1234)
    print "Finished creating database in %s seconds" % (time() - start)
    print "Giving the server index some time for a auto refresh"
    #wheelOfTime(reactor, 15)
    print "Finished creating database"

if __name__ == '__main__':
    from sys import argv
    if not '--fast' in argv:
        system('rm -rf ' + integrationTempdir)
        makedirs(inboxDirectory)
        makedirs(doneDirectory)
    port = randint(50000,60000)
    skipReindexTest = False
    reactor = Reactor()
    observableHttpServer = ObservableHttpServer(reactor, port)
    sruUpdateServer = observableHttpServer

    def ensureDirectory(dirname):
        fullpath = join(integrationTempdir, dirname)
        if not isdir(fullpath):
            makedirs(fullpath)
        return fullpath
    inboxDirectory = ensureDirectory('inboxDirectory')
    doneDirectory = ensureDirectory('doneDirectory')

    inbox = Inbox(reactor=reactor, inboxDirectory=inboxDirectory, doneDirectory=doneDirectory)

    server = be(serverMain(reactor,
        observableHttpServer,
        sruUpdateServer,
        inbox,
        baseURL='localhost',
        portNumber=port,
        databasePath=integrationTempdir,
        configDict={
            'defaultSortKeys': 'sort.mods.originInfo.dateIssued,,1',
            'applicationName': 'IntegrationTest BZV',
            'inboxDirectory': inboxDirectory,
            'doneDirectory': doneDirectory,
        }))
    #server.printTree()
    server.once.observer_init()
    if '--noReindex' in argv:
        skipReindexTest = True
        print 'Skipping testReindex!!'
        argv.remove('--noReindex')
    if '--fast' in argv and isdir(integrationTempdir):
        argv.remove('--fast')
        print "Reusing database in", integrationTempdir
    else:
        createDatabase()
    main()

